(* This file defines an abstract evaluator for Simple_java_Syntax *)

open Apron
       
open Simple_java_syntax  
open Simple_java_translate


module IntMap = Map.Make(struct type t=int let compare = ( - ) end)


(*Some useful functions to construct Apron's environment*)
let rec map_filter_rec f l1 l2 =
  match l1 with
  | [] -> List.rev l2
  | h1::t1 ->
    (match f h1 with
    | Some x -> map_filter_rec f t1 (x::l2)
    | None -> map_filter_rec f t1 l2)

let map_filter f l1 = map_filter_rec f l1 []

let collect_variables (prog:s_program) =
  List.flatten
    (List.map
      (fun classe -> map_filter
        (fun decl ->
          match decl with
          | Sd_var v -> (* We only collect integer variables *)
	     if (fst v).s_var_type <> St_int then
	       failwith "Only integer declarations are supported";
	     Some (classe.s_class_name, v)
          | Sd_function _ -> None) classe.s_class_body)
    prog);;

let print_variables l =
  List.iter
    (fun (class_name, var_decl) ->
      Printf.printf "%s.%s\n" class_name (fst var_decl).s_var_name) l;;

let get_main_procedure (prog: s_program) =
  let main = ref None in
  try
      List.iter (fun classe ->
        List.iter (fun decl ->
          match decl with
          | Sd_function f when f.s_proc_name = "main" ->
              main := Some f; raise Exit
          | Sd_function _ | Sd_var _ -> ()
        ) classe.s_class_body;
      ) prog;
      None
  with Exit -> !main;;

(* syntax of sexpr is in simple_java_syntax.ml[i] *)
let rec construct_texpr theenv themap ((sexpr, extent) : s_expr_e) : Texpr1.t =
  match sexpr with
  | Se_const(ct) -> failwith "construct_texpr : TODO"
  | Se_random(i1,i2) -> failwith "construct_texpr : TODO"
  | Se_var(sv) -> failwith "construct_texpr : TODO"
  | Se_unary(uop,se1) -> failwith "construct_texpr : TODO"
  | Se_binary(bop,se1,se2) -> failwith "construct_texpr : TODO"
  | Se_arrayaccess _  -> failwith "not yet implemented"


(* Apply a boolean expression (or its negation if invertflag is false) to a given abstract value. *)				  
let rec apply_boolexpr  themap (pre_state: 'a Abstract1.t) (invertflag:bool) ((sexpr, extent) : s_expr_e) = 
 let man = Abstract1.manager pre_state
 and env = Abstract1.env pre_state in
 match sexpr with
  | Se_const(ct) -> failwith "apply_boolexpr : TODO"
  | Se_random(i1,i2) -> failwith "Se_Random in boolexpr"
  | Se_var(sv) -> failwith "Se_var in boolexpr"
  | Se_unary(uop,se1) -> failwith "apply_boolexpr : TODO"
  | Se_binary(bop,se1,se2) ->failwith "apply_boolexpr : TODO"
  | Se_arrayaccess _  -> failwith "apply_boolexpr : not yet implemented";;


  
(*Execution of stmts / lists of stmts*)
				  
let rec execute_block variable_map (check: bool) (pre_state: 'a Abstract1.t) (block: s_block) =
  List.fold_left (execute_stmt variable_map check) pre_state block

and execute_stmt variable_map (check: bool) (pre_state : 'a Abstract1.t) ((stmt, extent) : s_command_e) =
  if check then
    begin
      Printf.printf "Abstract value : ";
      Abstract1.print Format.std_formatter pre_state;
      Printf.printf "\n";
    end;
  let env = Abstract1.env pre_state and man = Abstract1.manager pre_state in
  match stmt with
  | Sc_assign(svar,sexp) ->
     begin
       (if check then Format.printf "assign !\n"); 
       pre_state (*TODO : change here*)
     end
  | Sc_if(e,b1,b2)  -> (*TODO : change here*)
     begin 
       if check then Printf.printf "if!\n";
       pre_state 
     end
  | Sc_while(expr, block) -> (*TODO : change here*)
     begin 
       if check then Printf.printf "while!\n";
       pre_state 
     end
  | Sc_proc_call _ -> failwith "call: not yet implemented" 
  | Sc_assert _ -> failwith "assert: not yet implemented" 

let create_variable (class_name, (var_spec, initializers)) previous_map =
  let s = class_name ^ "." ^ var_spec.s_var_name in
  IntMap.add var_spec.s_var_uniqueId (Var.of_string s) previous_map

let create_variables l  =
  List.fold_right create_variable l IntMap.empty
		  
let abstract_run  (prog:s_program) =
  let manager = Polka.manager_alloc_loose () in
  let variables = collect_variables prog in
  let variable_map = create_variables variables in
  let env = Environment.make
    (Array.of_list
      (List.map (fun (id, v) -> v) (IntMap.bindings variable_map)))
    [| |] in
  let initial = Abstract1.top manager env in
  Abstract1.print Format.std_formatter initial;
    (match get_main_procedure prog with
    | Some proc -> execute_block variable_map true initial proc.s_proc_body
    | None -> failwith "no main function");;

(*Abstract evaluator main function*)
let abseval (prog:s_program) (javaenv: env) (fmt:Format.formatter) =
  Printf.printf "\n--------------------- Abstract Eval -----------------------\n";
  abstract_run prog; 
  Printf.printf "--------------------- end of Abstract Eval -----------------------\n";

